package nl.amis.spring.trial;
import java.util.HashMap;
import org.springframework.aop.framework.ProxyFactory;
import org.springframework.aop.target.HotSwappableTargetSource;

public class HotSwappeableTargets  {
  public HotSwappeableTargets() {
  }

  HashMap swapperCache = new HashMap(10);

  public Employee getEmployee() {
    Employee emp = new Employee();
    ProxyFactory pf = new ProxyFactory();
    
    // instead of setting the emp reference directly as target for the proxy,
    // we intervene with an additional HotSwappableTargetSource. The TargetSource  
    // implementation is asked for a target instance each time the AOP proxy handles a method invocation.
    // This essentially means that between invocations we can change the target
    // instance set on the TargetSource, so that subsequent TargetSource inspections
    // by the AOP Proxy Handler result in a different target object.
    HotSwappableTargetSource swapper = new HotSwappableTargetSource(emp);
    pf.setTargetSource(swapper);
    //pf.setInterfaces(new Class[]{Employee.class});
    pf.addAdvice(new EmployeeValidator());
    
    // add the swapper to a swapperCache; this allows us to later on retrieve
    // the swapper using a reference to the proxy, so that we can swap the POJO
    // underlying the proxy; note that pf.getProxy() is the reference that
    // consumers of this getEmployee() method will use instead of a direct
    // reference to an EmployeeImpl object.
    swapperCache.put(pf.getProxy(),swapper);
    return (Employee)pf.getProxy();
  }

  
  /**
   * This method will replace the reference to oldEmployee with a reference to newEmployee.
   * Anyone who held a reference to oldEmployee will still have a valid reference, as we
   * only swap the target for the proxy. All references to oldEmployee are in fact references
   * to a proxy for the oldEmployee. By changing the target for this proxy, all references
   * remain in tact and indirectly now reference the newEmployee
   * 
   * @param newEmployee
   * @param oldEmployee
   */
  private void swap(Employee oldEmployee, Employee newEmployee) {
     HotSwappableTargetSource swapper = (HotSwappableTargetSource)swapperCache.get(oldEmployee);     
     Object oldTarget = swapper.swap(newEmployee);
  }

  public static void main(String[] args) {
    HotSwappeableTargets mySecondAOP = new HotSwappeableTargets();
    // initialize a normal, non-AOPed bean emp:
    Employee emp = new Employee();
    emp.setFirstName("John");
    emp.setLastName("Doe");
    emp.setJobTitle("SALESMAN");
    emp.setSalary(new Float("1500"));

    // initialize a proxied bean, advised with a "Validator" Interceptor
    Employee scott = mySecondAOP.getEmployee();
    // note: scott is a proxy that implements the Employee interface. Internally
    // this proxy referes to an EmployeeImpl object that we never get to see directly
    scott.setFirstName("John");
    scott.setLastName("Smith");
    scott.setJobTitle("MANAGER");
    // we have advised scott with an advise that sets a default salary value for Managers:
    System.out.println("My dear employee "+scott.getFirstName()+" "+scott.getLastName()+" ("+scott.getJobTitle()+") = "+scott.getSalary());

    Employee alsoscott = scott; // alsoscott holds a reference to the same object (a proxy) that scott references

    // here we swap the EmployeeImpl underlying the scott reference with the emp EmployeeImpl
    mySecondAOP.swap(scott, emp);
    System.out.println("My dear employee "+scott.getFirstName()+" "+scott.getLastName()+" ("+scott.getJobTitle()+") = "+scott.getSalary());
    // all references to the object scott used to reference are still valid. That mains: alsoscott refers to the same employee as scott
    System.out.println("Also  "+alsoscott.getFirstName()+" "+alsoscott.getLastName()+" ("+alsoscott.getJobTitle()+") = "+alsoscott.getSalary());

    // setting a property on emp now also means setting it on scott - as scott is a reference to a proxy for the same object that emp references
    emp.setSalary(new Float("2500"));
    System.out.println("My dear employee "+scott.getFirstName()+" "+scott.getLastName()+" ("+scott.getJobTitle()+") = "+scott.getSalary());
    // scott was advised with a validator; although a swap has taken place and scott is now associated with another Employee object (or should we say
    // another EmployeeImpl object), this advise is still in place
    scott.setSalary(new Float("5000.32"));
    System.out.println("The new salary for "+scott.getFirstName()+" "+scott.getLastName()+" ("+scott.getJobTitle()+") = "+scott.getSalary());
  }
}
